/*
 * Copyright E. Bigeon (2015)
 *
 * emmanuel@bigeon.fr
 *
 * This software is a computer program whose purpose is to
 * provide a swt window for console applications.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
/**
 * gclc-swt:fr.bigeon.gclc.swt.SWTConsole.java
 * Created on: Apr 18, 2015
 */
package fr.bigeon.gclc.swt;

import java.io.IOException;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.FocusAdapter;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.KeyAdapter;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;

import fr.bigeon.collections.ArrayRibbon;
import fr.bigeon.collections.Ribbon;
import fr.bigeon.gclc.ConsoleApplication;
import fr.bigeon.gclc.manager.ConsoleManager;

/** A SWT component to connect to gclc {@link ConsoleApplication}
 * <p>
 *
 * @author Emmanuel Bigeon */
public class SWTConsole extends Composite implements ConsoleManager {
    /** A key listener to validate commands and manage the history of commands
     *
     * @author Emmanuel Bigeon */
    public static final class HistoryTextKeyListener extends KeyAdapter {

        /** The size of commands history */
        private static final int DEFAULT_HISTORY_SIZE = 10;
        /** The history ribbon */
        private final Ribbon<String> commands;
        /** The current index in history search */
        private int currentIndex = 0;
        /** The console to write the commands in */
        private final Text consoleInput;
        /** The console to notify of command validation */
        private final SWTConsole console;

        /** @param console the console
         * @param consoleInput the text to write commands in */
        public HistoryTextKeyListener(SWTConsole console, Text consoleInput) {
            super();
            this.console = console;
            this.consoleInput = consoleInput;
            this.commands = new ArrayRibbon<>(DEFAULT_HISTORY_SIZE);
        }

        @Override
        public void keyPressed(KeyEvent e) {
            // Enter validates the command if prompting
            if (e.keyCode == '\r') {
                commands.add(consoleInput.getText());
                console.validateInput();
                currentIndex = -1;
            }

            // Upper arrow retrieves previous commands
            if (e.keyCode == SWT.ARROW_UP &&
                currentIndex < commands.size() - 1) {
                currentIndex++;
                consoleInput.setText(
                        commands.get(commands.size() - currentIndex - 1));
                consoleInput.setSelection(consoleInput.getText().length());
            }

            // Lower arrow retrieves next commands
            if (e.keyCode == SWT.ARROW_DOWN) {
                if (currentIndex == 0) {
                    currentIndex--;
                    consoleInput.setText(new String());
                } else if (currentIndex > 0) {
                    consoleInput.setText(commands
                            .get(commands.size() - (--currentIndex) - 1));
                    consoleInput.setSelection(consoleInput.getText().length());
                }
            }
        }
    }

    /**
     *
     */
    private static final int LAYOUT_NB_COLUMNS = 2;
    /** The cmd prefix in the output console */
    private static final String CMD_PREFIX = "[CMD] "; //$NON-NLS-1$
    /** The console output text field */
    private final Text consoleOutput;
    /** The console input text field */
    private final Text consoleInput;
    /** The prompt label */
    private final Label lblPromptlabel;
    /** The prompt text */
    private String prompt = ">"; //$NON-NLS-1$
    /** The command entered by the user */
    private String command = null;
    /** If the prompt should be active */
    private boolean prompting = false;
    /** The object for thread synchronization with the prompt */
    private final Object promptLock = new Object();

    /** Create the composite.
     *
     * @param parent the prent composite
     * @param style the composite style */
    public SWTConsole(Composite parent, int style) {
        super(parent, style);

        setLayout(new GridLayout(LAYOUT_NB_COLUMNS, false));

        consoleOutput = new Text(this, SWT.BORDER | SWT.READ_ONLY | SWT.WRAP |
                                       SWT.V_SCROLL | SWT.MULTI);
        consoleOutput.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true,
                LAYOUT_NB_COLUMNS, 1));
        consoleOutput.setRedraw(true);
        consoleOutput.addFocusListener(new FocusAdapter() {
            @SuppressWarnings("synthetic-access")
            @Override
            public void focusGained(FocusEvent e) {
                consoleInput.setFocus();
            }
        });

        lblPromptlabel = new Label(this, SWT.NONE);
        lblPromptlabel.setText(prompt);

        consoleInput = new Text(this, SWT.BORDER);
        consoleInput.setLayoutData(
                new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));
        consoleInput.addKeyListener(
                new HistoryTextKeyListener(this, consoleInput));

    }

    /**
     * 
     */
    protected void validateInput() {
        if (prompting) {
            synchronized (promptLock) {
                command = consoleInput.getText();
                prompting = false;
                consoleInput.setText(new String());
                consoleInput.setEnabled(false);
                consoleOutput
                        .append(CMD_PREFIX + command + System.lineSeparator());
                promptLock.notifyAll();
            }
        }
    }

    @Override
    protected void checkSubclass() {
        // Disable the check that prevents subclassing of SWT components
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.ConsoleManager#getPrompt() */
    @Override
    public String getPrompt() {
        return prompt;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.ConsoleManager#print(java.lang.String) */
    @Override
    public void print(final String text) {
        Display.getDefault().syncExec(new Runnable() {
            @SuppressWarnings("synthetic-access")
            @Override
            public void run() {
                consoleOutput.append(text);
            }
        });
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.ConsoleManager#println() */
    @Override
    public void println() {
        Display.getDefault().syncExec(new Runnable() {
            @SuppressWarnings("synthetic-access")
            @Override
            public void run() {
                consoleOutput.append(System.lineSeparator());
            }
        });
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.ConsoleManager#println(java.lang.String) */
    @Override
    public void println(final String message) {
        Display.getDefault().syncExec(new Runnable() {
            @SuppressWarnings("synthetic-access")
            @Override
            public void run() {
                consoleOutput.append(message + System.lineSeparator());
            }
        });

    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.ConsoleManager#prompt() */
    @Override
    public String prompt() {
        synchronized (promptLock) {
            try {
                Display.getDefault().syncExec(new Runnable() {
                    @SuppressWarnings("synthetic-access")
                    @Override
                    public void run() {
                        consoleInput.setEnabled(true);
                        consoleInput.setFocus();
                    }
                });
                prompting = true;
                promptLock.wait();
            } catch (final InterruptedException e) {
                command = null;
            }
        }
        return command;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.ConsoleManager#prompt(java.lang.String) */
    @Override
    public String prompt(final String message) {
        synchronized (promptLock) {
            try {
                Display.getDefault().syncExec(new Runnable() {
                    @SuppressWarnings("synthetic-access")
                    @Override
                    public void run() {
                        lblPromptlabel.setText(message);
                        // relayout
                        SWTConsole.this.layout();
                        consoleInput.setEnabled(true);
                        consoleInput.setFocus();
                    }
                });
                prompting = true;
                promptLock.wait();
            } catch (final InterruptedException e) {
                command = null;
            } finally {
                Display.getDefault().syncExec(new Runnable() {
                    @SuppressWarnings("synthetic-access")
                    @Override
                    public void run() {
                        lblPromptlabel.setText(prompt);
                        lblPromptlabel.pack();
                    }
                });
            }
        }
        return command;
    }

    /* (non-Javadoc)
     * @see org.eclipse.swt.widgets.Composite#setFocus() */
    @Override
    public boolean setFocus() {
        return consoleInput.setFocus();
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.ConsoleManager#setPrompt(java.lang.String) */
    @Override
    public void setPrompt(final String prompt) {
        this.prompt = prompt;
        Display.getDefault().syncExec(new Runnable() {
            @SuppressWarnings("synthetic-access")
            @Override
            public void run() {
                lblPromptlabel.setText(prompt);
                lblPromptlabel.pack();
            }
        });
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleManager#close() */
    @Override
    public void close() throws IOException {
        consoleInput.setEnabled(false);
        consoleOutput.setEnabled(false);
    }

}
